
-- Copyright (C) by Jiancong Ni
local tracker = require("resty.fastdfs.tracker")
local storage = require("resty.fastdfs.storage")
local utils = require('resty.fastdfs.utils')

local var = ngx.var
local split_fileid = utils.split_fileid
local BUFFER_SIZE = 1024 * 1024 * 1
local tonumber = tonumber
local io = io

module(...)

local _M = { _VERSION = '0.01' }
local mt = { __index = _M }

local function getExtension(str) return str:match(".+%.(%w+)$") end

local function get_address(upstream)
  var.GP_upstream = upstream
  local port = 0
  local address = var.GP_address
  if address and address:byte(1)~=117 then
    local flag = address:find(':', 8)
    port = tonumber(address:sub(flag+1, -1))
    address = address:sub(1, flag-1)
    return address, port
  end
  return address
end

local function fastdfs_get_storage(etc, upstream, method, fileid)
	local ok, qss
	local up = upstream or "up_fastdfs_traker"
	local err = "failed get_address from " .. up
	local ip, port = get_address(up)
	if ip and port then 
		local tk = tracker:new()
		tk:set_timeout(3000)
		ok, err = tk:connect({host=ip, port=port})
		if ok then
			if method == "update" then
				local group_name, file_name
				group_name, file_name, err = split_fileid(fileid or "")
				if group_name and file_name then
					qss, err = tk:query_storage_update(group_name, file_name)
				end
			elseif method == "fetch" then
				local group_name, file_name
				group_name, file_name, err = split_fileid(fileid or "")
				if group_name and file_name then
					qss, err = tk:query_storage_fetch(group_name, file_name)
				end
			else
				qss, err = tk:query_storage_store(fileid)
			end
			tk:set_keepalive(300000, 30)
			if qss then return qss.host, nil, qss end
		end
	end
	return nil, err
end

local function fastdfs_check_active(etc, group, upstream)
	local host, err = fastdfs_get_storage(etc, upstream, nil, group)
	return host, err
end

local function fastdfs_upload(etc, file, no_local_first, upstream)
	local host, err, qss, fp, res, res1, ok
	local ext = getExtension(file)
	fp, err = io.open(file, "rb")
	if fp then
		local st = storage:new()
		st:set_timeout(600000)
		if not no_local_first then
			qss = {
				host = "127.0.0.1",
				port = "23000",
				group_name = nil,
				store_path_index = 255
			}
			ok, err = st:connect(qss)
		end
		if not ok then
			host, err, qss = fastdfs_get_storage(etc, upstream)
			if qss then ok, err = st:connect(qss) end
		end
		if ok then
			local buff = fp:read(BUFFER_SIZE) or ""
			res1, err = st:upload_appender_by_buff(buff, ext)
			if res1 then
				res = true
				buff = fp:read(BUFFER_SIZE)
				while buff do
					res, err = st:append_by_buff(res1.group_name, res1.file_name, buff)
					if not res then
						st:delete_file(res1.group_name, res1.file_name)
						break
					end
					buff = fp:read(BUFFER_SIZE)
				end
			end
			st:set_keepalive(300000, 30)
		end
		fp:close()
	end
	if res then return res1.group_name .. "/" .. res1.file_name end
	return nil, err
end

local function fastdfs_upload_buf(buf, ext, appender, no_local_first, upstream)
	local host, err, qss, res, ok
	local st = storage:new()
	st:set_timeout(600000)
	if not no_local_first then
		qss = {
			host = "127.0.0.1",
			port = "23000",
			group_name = nil,
			store_path_index = 255
		}
		ok, err = st:connect(qss)
	end
	if not ok then
		host, err, qss = fastdfs_get_storage(etc, upstream)
		if qss then ok, err = st:connect(qss) end
	end
	if ok then
		if appender then
			res, err = st:upload_appender_by_buff(buf, ext)
		else
			res, err = st:upload_by_buff(buf, ext)
		end
		st:set_keepalive(300000, 30)
	end
	if res then return res.group_name .. "/" .. res.file_name end
	return nil, err
end

local function fastdfs_append(fileid, file, upstream)
	local host, err, qss, fp, res, res1, ok
	local ext = getExtension(file)
	fp, err = io.open(file, "rb")
	if fp then
		local st = storage:new()
		st:set_timeout(600000)
		host, err, qss = fastdfs_get_storage(etc, upstream, "update", fileid)
		if qss then ok, err = st:connect(qss) end
		if ok then
			local buff = fp:read(BUFFER_SIZE) or ""
			res1, err = st:append_by_buff1(fileid, buff)
			if res1 then
				res = true
				buff = fp:read(BUFFER_SIZE)
				while buff do
					res, err = st:append_by_buff1(fileid, buff)
					if not res then
						break
					end
					buff = fp:read(BUFFER_SIZE)
				end
			end
			st:set_keepalive(300000, 30)
		end
		fp:close()
	end
	if res then return "ok" end
	return nil, err
end

local function fastdfs_append_buf(fileid, buf, upstream)
	local host, err, qss, res, ok
	local st = storage:new()
	st:set_timeout(600000)
	host, err, qss = fastdfs_get_storage(etc, upstream, "update", fileid)
	if qss then ok, err = st:connect(qss) end
	if ok then
		res, err = st:append_by_buff1(fileid, buf)
		st:set_keepalive(300000, 30)
	end
	if res then return "ok" end
	return nil, err
end

local function fastdfs_modify_buf(fileid, buf, offset, upstream)
	local host, err, qss, res,ok
	local st = storage:new()
	st:set_timeout(600000)
	host, err, qss = fastdfs_get_storage(etc, upstream, "update", fileid)
	if qss then ok, err = st:connect(qss) end
	if ok then
		res, err = st:modify_by_buff1(fileid, buf, offset)
		st:set_keepalive(300000, 30)
	end
	if res then return "ok" end
	return nil, err
end

local function fastdfs_delete(etc, fileid, upstream)
	local ok, qss
	local up = upstream or "up_fastdfs_traker"
	local err = "failed get_address from " .. up
	local ip, port = get_address(up)
	if ip and port then 
		local tk = tracker:new()
		tk:set_timeout(3000)
		ok, err = tk:connect({host=ip, port=port})
		if ok then
			qss, err = tk:query_storage_fetch1(fileid)
			tk:set_keepalive(300000, 30)
			if qss then
				local st = storage:new()
				st:set_timeout(3000)
				ok, err = st:connect(qss)
				if ok then return st:delete_file1(fileid) end
			end
		end
	end
	return nil, err
end

_M.fastdfs_upload = fastdfs_upload
_M.fastdfs_get_storage = fastdfs_get_storage
_M.fastdfs_delete = fastdfs_delete
_M.fastdfs_check_active = fastdfs_check_active
_M.fastdfs_upload_buf = fastdfs_upload_buf
_M.fastdfs_append = fastdfs_append
_M.fastdfs_append_buf = fastdfs_append_buf
_M.fastdfs_modify_buf = fastdfs_modify_buf

return _M

